(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Narrowing_Impl
imports
  Innermost_Usable_Rules_Impl 
  Dependency_Graph_Impl
  Narrowing
begin

definition qnarrows_impl :: "(('f,string)term \<Rightarrow> bool) \<Rightarrow> ('f,string)rules \<Rightarrow> ('f,string)term \<Rightarrow> (('f,string)term \<times> ('f,string)subst) list"
  where "qnarrows_impl isnf R t \<equiv> concat (map (\<lambda> p. let tp = t |_ p in
  if is_Fun tp then [(replace_at (t \<cdot> \<mu>1) p (r \<cdot> \<mu>2), \<mu>1). (l,r) \<leftarrow> R, (\<mu>1,\<mu>2) \<leftarrow> option_to_list (mgu_var_disjoint_string tp l), (\<forall> la \<in> set (args (l \<cdot> \<mu>2)). isnf la)] else []) (poss_list t))"

lemma qnarrows_impl[simp]: 
  "set (qnarrows_impl (\<lambda> t. t \<in> NF_terms Q) R t) = {(t',\<mu>). \<exists> r p. (t,t') \<in> qnarrows_r_p_s Q (set R) r p \<mu>}" (is "?l = ?r")
proof -
  note d = qnarrows_impl_def[unfolded Let_def]  qnarrows_r_p_s_def
  note sc = set_concat set_map split
  {
    fix t' \<mu>
    assume "(t',\<mu>) \<in> ?r"
    from this[unfolded d] obtain lr p \<mu>2 where
      p: "p \<in> poss t" and
      tp: "is_Fun (t |_ p)" and
      R: "lr \<in> set R" and
      mgu: "mgu_var_disjoint_string (t |_ p) (fst lr) = Some (\<mu>,\<mu>2)" and
      NF: "set (args (fst lr \<cdot> \<mu>2)) \<subseteq> NF_terms Q" and
      t': "t' = replace_at (t \<cdot> \<mu>) p (snd lr \<cdot> \<mu>2)" by blast
    from tp have tp: "\<And> r q. (if is_Fun (t |_ p) then r else q) = r" by simp
    obtain l r where lr: "lr = (l,r)" by force
    have "(t',\<mu>) \<in> ?l" 
      unfolding d sc
      by (rule, unfold poss_list_sound, rule imageI[OF p], unfold tp sc,
      rule, rule imageI[OF R[unfolded lr]], unfold sc, rule, rule imageI[of "(\<mu>,\<mu>2)"],
      insert mgu NF t' lr, auto)
  }
  hence "?r \<subseteq> ?l" by blast
  moreover
  {
    fix t' \<mu>
    assume "(t',\<mu>) \<in> ?l"
    from this[unfolded d sc]
    obtain p l r \<mu>2 where 
      p: "p \<in> poss t" and tp: "is_Fun (t |_ p)" and
      lr: "(l,r) \<in> set R" and
      mgu: "mgu_var_disjoint_string (t |_ p) l = Some (\<mu>,\<mu>2)" and
      NF: "\<forall> t \<in> set (args (l \<cdot> \<mu>2)). t \<in> NF_terms Q" and
      t': "t' = replace_at (t \<cdot> \<mu>) p (r \<cdot> \<mu>2)" by auto
    have "(t',\<mu>) \<in> ?r"
      unfolding d
      by (rule, unfold sc, rule exI[of _ "(l,r)"], rule exI[of _ p], rule, unfold sc, rule exI[of _ \<mu>2],
      insert p tp lr mgu NF t', auto)
  } hence "?l \<subseteq> ?r" by blast
  ultimately show ?thesis by blast
qed

definition
  narrowing_proc ::
    "('dpp, 'f :: {linorder,show}, string) dpp_ops \<Rightarrow> 
     ('f,string)rule \<Rightarrow> pos \<Rightarrow> ('f,string)rules \<Rightarrow> 'dpp proc"
where
  "narrowing_proc I st p sts dpp \<equiv> 
    check_return (do {
     let (s,t) = st;
     check (dpp_ops.NFQ_subset_NF_rules I dpp \<or> dpp_ops.Q I dpp = [] \<and> linear_term t) (shows ''innermost or full rewriting required (and linearity of t in full rewriting case)'');
     let ic = icap_impl_dpp_mv I dpp;
     let isnf = dpp_ops.is_QNF I dpp;
     let pairs = dpp_ops.pairs I dpp;
     check (p \<in> set (poss_list (ic [s] t))) (shows ''position not contained in capped term of '' +@+ shows t);
     let tp = t |_ p;
     let narrows = qnarrows_impl isnf (dpp_ops.rules I dpp) tp;
     let sts' = map (\<lambda> (t',\<mu>). (s \<cdot> \<mu>, replace_at (t \<cdot> \<mu>) p t')) narrows;
     check_allm (\<lambda> new. check (\<exists> st' \<in> set sts. instance_rule new st')
                (shows ''could not find narrowed pair '' +@+ shows_rule new)) sts';
     let iedg = is_iedg_edge_dpp I dpp (s,t);
     check_subseteq (vars_term_list tp) (vars_term_list s) <+? (\<lambda> x. shows (''variable '' @ x @ '' only occurs on rhs of pair''));
     check (st \<in> set (dpp_ops.P I dpp) \<or> dpp_ops.R I dpp = []) (shows ''strict DP or no strict rules required'');
     (if isnf tp then check_allm (\<lambda> (u,v). 
         do {
             check (p \<in> set (poss_list u)) (shows ''position not contained in lhs of pair '' +@+ shows_rule (u,v));
             case mgu_var_disjoint_string tp (u |_ p) of None \<Rightarrow> OK | Some (\<mu>1,\<mu>2) \<Rightarrow> check (\<not> isnf (s \<cdot> \<mu>1) \<or> \<not> isnf (u \<cdot> \<mu>2)) (shows ''t |_ p and u |_ p unify and satisfy variable condition for pair (u,v) = ''
                +@+ shows_rule (u,v))
         }) (filter (\<lambda> (u,v). iedg u) pairs) 
       else OK)
   })
   (dpp_ops.replace_pair I dpp st sts)"

lemma narrowing_proc: assumes I: "dpp_spec I"
 shows "dpp_spec.sound_proc_impl I (narrowing_proc I (st :: ('f :: {linorder,show},string)rule)
  p sts)"
proof -
  from assms interpret dpp_spec I .
  show ?thesis
  proof
    fix d d'    
    assume ok: "narrowing_proc I st p sts d = Inr d'" and fin: "finite_dpp (dpp d')"
    let ?P = "set (P d)"
    let ?Pw = "set (Pw d)"
    let ?Q = "set (Q d)"
    let ?R = "set (R d)"
    let ?Rw = "set (Rw d)"
    let ?Rb = "?R \<union> ?Rw"
    let ?Rb' = "set (rules d)" 
    let ?Pb = "?P \<union> ?Pw"
    let ?Pb' = "set (pairs d)"
    obtain s t where st: "st = (s,t)" by force
    note ok = ok[unfolded narrowing_proc_def st icap_impl_dpp_icap_mv[OF I] Let_def dpp_spec_sound, simplified] 
    from ok have d': "d' = replace_pair d (s,t) sts" by auto
    have d': "dpp d' = (replace (s,t) (set sts) ?P, replace (s,t) (set sts) ?Pw, ?Q, ?R, ?Rw)" 
      unfolding d' unfolding replace_pair_sound by simp
    note fin = fin[unfolded this]
    note ok_narrow = ok[THEN conjunct2, THEN conjunct2, THEN conjunct1]
    note ok_precond = ok[THEN conjunct2, THEN conjunct2, THEN conjunct2, THEN conjunct2, THEN conjunct2, THEN conjunct1]
    show "finite_dpp (dpp d)" unfolding dpp_spec_sound
    proof (rule narrowing_proc[OF fin _ _ _ _ _ _ icap])
      from ok show "(s,t) \<in> ?P \<or> ?R = {}" by auto
    next
      from ok show "linear_term t \<and> ?Q = {} \<or> NF_terms ?Q \<subseteq> NF_trs ?Rb" by auto
    next
      from ok show "vars_term t |_ p \<subseteq> vars_term s" by auto
    next
      from ok have "p \<in> poss (icap_mv ?Rb ?Q {s} t)" by simp
      thus "p \<in> poss (icap' ?Rb ?Q {mv_xvar s} (mv_xvar t))" unfolding icap_mv_def using icap'[of ?Rb ?Q "{s}" t] by auto
    next      
      fix u v
      assume tp: "t |_ p \<in> NF_terms ?Q" and edge: "((s,t),(u,v)) \<in> DG ?Pb ?Q ?Rb"
      from edge[unfolded DG_def] obtain \<sigma> \<tau> where steps: "(t \<cdot> \<sigma>, u \<cdot> \<tau>) \<in> (qrstep ?Q ?Rb)^*"
        and NFs: "s \<cdot> \<sigma> \<in> NF_terms ?Q" and NFu: "u \<cdot> \<tau> \<in> NF_terms ?Q" 
        and SN: "SN_on (qrstep ?Q ?Rb) {t \<cdot> \<sigma>}"
        and mem: "(u,v) \<in> ?Pb" by auto
      have edge: "is_iedg_edge_dpp I d (s,t) u"
        by (rule is_iedg_edge_dpp_sound[OF I NFs NFu], insert SN steps, auto)
      hence edge: "is_iedg_edge_dpp I d (s,t) u = True" by simp
      from tp have tp: "t |_ p \<in> NF_terms ?Q = True" by simp
      note ok = ok_precond[unfolded tp, simplified, rule_format, of u v, unfolded edge]
      from ok mem have pu: "p \<in> poss u" by simp
      show "p \<in> poss u \<and> (\<forall> \<mu>1 \<mu>2. mgu_var_disjoint_string (t |_ p) (u |_ p) = Some (\<mu>1,\<mu>2) \<longrightarrow> \<not> {s \<cdot> \<mu>1, u \<cdot> \<mu>2} \<subseteq> NF_terms ?Q)"
      proof (rule conjI[OF pu], intro allI impI)
        fix \<mu>1 \<mu>2
        assume mgu: "mgu_var_disjoint_string (t |_ p) (u |_ p) = Some (\<mu>1, \<mu>2)"
        from ok[unfolded mgu] mem show "\<not> {s \<cdot> \<mu>1, u \<cdot> \<mu>2} \<subseteq> NF_terms ?Q" by auto
      qed
    next
      fix rule q t' \<mu>
      assume narr: "(t |_ p, t') \<in> qnarrows_r_p_s ?Q ?Rb rule q \<mu>"
      obtain l r where rule: "rule = (l,r)" by force
      from ok_narrow[rule_format, of t' \<mu>] narr[unfolded rule] 
      show "\<exists> st' \<in> set sts. instance_rule (s \<cdot> \<mu>, replace_at (t \<cdot> \<mu>) p t') st'" by auto
    qed
  qed
qed


end