theory Seq
imports Main
begin

text {*The type of infinite sequences.*}
type_synonym 'a seq = "nat \<Rightarrow> 'a"

abbreviation
  cons :: "'a \<Rightarrow> 'a seq \<Rightarrow> 'a seq" (infixr "#s" 65)
where
  "x #s S \<equiv> (\<lambda>i. case i of 0 \<Rightarrow> x | Suc n \<Rightarrow> S n)"

text {*The type of partial sequences.*}
type_synonym 'a pseq = "'a option seq"

abbreviation (input) linked :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow>'a seq \<Rightarrow> bool" where
  "linked P S \<equiv> \<forall>i. P (S i) (S (Suc i))"

abbreviation (input) linkedrel :: "('a \<times> 'a) set \<Rightarrow> 'a seq \<Rightarrow> bool" where
  "linkedrel r S \<equiv> linked (\<lambda>x y. (x, y) \<in> r) S"

lemma cons_linked:
  assumes "P x (S 0)" and "linked P S"
  shows "linked P (x #s S)" (is "linked P ?S")
proof
  fix i
  from assms show "P (?S i) (?S (Suc i))" by (cases i) simp_all
qed

lemma cons_linkedrel:
  assumes "(x, S 0) \<in> r" and "linkedrel r S"
  shows "linkedrel r (x #s S)"
  using cons_linked[of "\<lambda>x y. (x, y) \<in> r", OF assms] .

(*TODO: subsumes Abstract_Rewriting.iseq_imp_condensed_iseq*)
lemma seq_imp_relpow:
  assumes "linkedrel r S" shows "(S i, S (i + j)) \<in> r^^j"
proof (induct "i + j" arbitrary: j)
  case (Suc n) thus ?case using assms by (cases j) auto
qed simp

(*TODO: subsumes Abstract_Rewriting.iseq_imp_steps*)
lemma seq_imp_rtrancl:
  assumes "linkedrel r S" and "i \<le> j" shows "(S i, S j) \<in> r^*"
proof -
  from assms(2) obtain n where "j = i + n" by (induct "j - i" arbitrary: j) force+
  with seq_imp_relpow[OF assms(1), of i n] show ?thesis by (simp add: relpow_imp_rtrancl)
qed

(*TODO: subsumes Termination_Inheritance.stepfun_imp_seq*)
lemma stepfun_imp_seq:
  assumes "\<forall>i\<ge>n::nat. f i > i \<and> P (S i) (S (f i))"
  shows "linked P (\<lambda>i. S ((f ^^ i) n))"
    (is "linked P ?T")
proof
  fix i
  from assms have "(f ^^ i) n \<ge> n" by (induct i) auto
  with assms[THEN spec[of _ "(f ^^ i) n"]]
    show "P (?T i) (?T (Suc i))" by simp
qed

(*TODO: subsumes Termination_Inheritance.steps_imp_seq*)
lemma steps_imp_seq:
  assumes "\<forall>i\<ge>n::nat. \<exists>j>i. P (S i) (S j)" shows "\<exists>T. linked P T"
proof -
  from assms have "\<forall>i\<in>{i. i \<ge> n}. \<exists>j>i. P (S i) (S j)" by auto
  from bchoice[OF this]
    obtain f where seq: "\<forall>i\<ge>n. f i > i \<and> P (S i) (S (f i))" by auto
  from stepfun_imp_seq[of n f P S, OF this] show ?thesis by fast
qed

lemma seq_imp_trancl:
  assumes "linkedrel r S" and "i < j" shows "(S i, S j) \<in> r^+"
proof -
  from assms(2) obtain n where "j = i + Suc n" by (induct "j - i" arbitrary: j) force+
  with seq_imp_relpow[OF assms(1), of i "Suc n"]
    show ?thesis unfolding trancl_power by force
qed

end